from typing import Any, List, Mapping, Optional, Union

from typing_extensions import Literal, NotRequired, TypedDict

FieldType = Union[
    Literal["section"],
    Literal["radiobutton"],
    Literal["dropdown"],
    Literal["checkbox"],
    Literal["nric"],
    Literal["email"],
    Literal["table"],
    Literal["number"],
    Literal["rating"],
    Literal["yes_no"],
    Literal["decimal"],
    Literal["textfield"],  # short text
    Literal["textarea"],  # long text
    Literal["attachment"],
    Literal["date"],
    Literal["mobile"],
    Literal["homeno"],
]

FormField = TypedDict(
    "FormField",
    {
        "_id": str,
        "question": str,
        "fieldType": FieldType,
        "isHeader": bool,
        "signature": Optional[str],
        "answer": Optional[str],
        "answerArray": Optional[Union[List[str], List[List[str]]]],
    },
)
DecryptedContent = TypedDict(
    "DecryptedContent",
    {"responses": List[FormField], "verified": NotRequired[Mapping[str, Any]]},
)

DecryptedFile = TypedDict("DecryptedFile", {"filename": str, "content": bytes})
DecryptedAttachments = Mapping[str, DecryptedFile]
DecryptedContentAndAttachments = TypedDict(
    "DecryptedContentAndAttachments",
    {"content": DecryptedContent, "attachments": DecryptedAttachments},
)


EncryptedAttachmentRecords = Mapping[str, str]

EncryptedFileContent = TypedDict(
    "EncryptedFileContent",
    {"submission_public_key": str, "nonce": str, "binary": bytes},
)


DecryptParams = TypedDict(
    "DecryptParams",
    {
        "encryptedContent": str,
        "version": str,
        "verifiedContent": Optional[str],
        "attachmentDownloadUrls": Optional[EncryptedAttachmentRecords],
    },
)



from typing import Any, Dict


def _filter_form_fields(field: Dict[str, Any]):
    return (
        (
            isinstance(field.get("ans"), str)
            or isinstance(field.get("answerArray"), list)
            or field.get('isHeader') is not None
        )
        and field.get("_id") is not None
        and field.get("fieldType") is not None
        and field.get("question") is not None
    )


def determine_is_form_fields(tbd: Any):
    if not isinstance(tbd, list):
        return False

    filtered = list(filter(lambda internal: _filter_form_fields(internal), tbd))
    return len(filtered) == len(tbd)



from typing_extensions import TypedDict

VerificationAuthenticateOptions = TypedDict(
    "VerificationAuthenticateOptions",
    {
        "signatureString": str,
        "submissionCreatedAt": int,
        "fieldId": str,
        "answer": str,
        "publicKey": str,
    },
)

VerificationSignatureSchema = TypedDict(
    "VerificationSignatureSchema", {"v": str, "t": int, "s": str, "f": str}
)

HeaderSignatureSchema = TypedDict(
    "HeaderSignatureSchema", {"v1": str, "t": int, "s": str, "f": str}
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['formsg', 'formsg.schemas', 'formsg.util']

package_data = \
{'': ['*']}

install_requires = \
['PyNaCl>=1.5.0,<2.0.0', 'pytest>=6.2.5,<7.0.0', 'requests>=2.27.1,<3.0.0']

extras_require = \
{':python_version < "3.10"': ['typing_extensions>=3.10.0,<4.0.0']}

setup_kwargs = {
    'name': 'formsg-python-sdk',
    'version': '0.1.9',
    'description': 'Python SDK for FormSG',
    'long_description': None,
    'author': 'Chin Ying',
    'author_email': 'chinying@open.gov.sg',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



import urllib
from typing import Optional

from formsg.exceptions import MissingSecretKeyException, WebhookAuthenticateException
from formsg.util.parser import parse_signature_header
from formsg.util.webhook import has_epoch_expired, is_signature_valid, sign


class Webhook(object):
    def __init__(self, public_key: str, secret_key: Optional[str] = None):
        self.public_key = public_key
        self.secret_key = secret_key

    def authenticate(self, header: str, uri: str) -> bool:
        """
        Injects the webhook public key for authentication
        :param header: X-FormSG-Signature header
        :param uri: The endpoint that FormSG is POSTing to
        :rtype: :class:`bool` true if the header is verified
        :raises WebhookAuthenticateException: If the signature or uri cannot be verified
        """
        signature_header = parse_signature_header(header)
        [signature, epoch, submission_id, form_id] = [
            signature_header["v1"],
            signature_header["t"],
            signature_header["s"],
            signature_header["f"],
        ]

        # verify signature authenticity
        if not is_signature_valid(uri, signature_header, self.public_key):
            raise WebhookAuthenticateException(
                f"Signature could not be verified for uri={uri} submission_id={submission_id} form_id={form_id} epoch={epoch} signature={signature}"
            )

        # verify epoch recency
        if has_epoch_expired(epoch):
            raise WebhookAuthenticateException(
                "Signature is not recent for uri={uri} submission_id={submission_id} form_id={form_id} epoch={epoch} signature={signature}"
            )

        return True

    def generate_signature(self, params) -> str:
        """
        Generates a signature based on the URI, submission ID and epoch timestamp.
        :param params The parameters needed to generate the signature
        :param params.uri Full URL of the request
        :param params.submissionId Submission Mongo ObjectId saved to the database
        :param params.epoch Number of milliseconds since Jan 1, 1970
        :returns the generated signature
        :raises MissingSecretKeyException if a secret key is not provided when instantiating this class
        :raises TypeError if any parameters are undefined
        """
        if not self.secret_key:
            raise MissingSecretKeyException()

        [submission_id, uri, form_id, epoch] = [
            params["submissionId"],
            params["uri"],
            params["formId"],
            params["epoch"],
        ]
        if not submission_id or not uri or not form_id or not epoch:
            raise TypeError(
                "submissionId, uri, formId, or epoch must be provided to generate a webhook signature"
            )

        parsed_url = urllib.parse.urlparse(uri).geturl()
        base_string = f"{parsed_url}.{submission_id}.{form_id}.{epoch}"
        return sign(base_string, self.secret_key).decode("utf-8")

    def construct_header(self, params) -> str:
        """
        Constructs the `X-FormSG-Signature` header
        :param params The parameters needed to construct the header
        :param params.epoch Epoch timestamp
        :param params.submissionId Mongo ObjectId
        :param params.formId Mongo ObjectId
        :param params.signature A signature generated by the generate_signature() function
        :rtype str: The `X-FormSG-Signature` header
        :raises MissingSecretKeyException if a secret key is not provided when instantiating this class
        """
        [epoch, submission_id, form_id, signature] = [
            params["epoch"],
            params["submissionId"],
            params["formId"],
            params["signature"],
        ]
        if not self.secret_key:
            raise MissingSecretKeyException()
        return f"t={epoch},s={submission_id},f={form_id},v1={signature}"



import base64
import json
import logging
from typing import Any, Dict, List, Mapping, Union

from nacl.exceptions import CryptoError
from nacl.public import Box, PrivateKey, PublicKey
from nacl.signing import VerifyKey

from formsg.schemas.crypto import EncryptedFileContent

logger = logging.getLogger(__name__)


def verify_signed_message(msg: bytes, public_key: str) -> Dict[str, Any]:
    """
    helper method to verify a signed message
    :param msg: message to verify
    :param public_key: the public key to authenticate the signed message with
    :returns the signed message if successful, else an error will be thrown
    raises Exception if mesasage cannot be verified
    """
    verify_key = VerifyKey(base64.b64decode(public_key))
    opened_message = verify_key.verify(msg)
    if not opened_message:
        raise Exception("Failed to open signed message with given public key")
    return json.loads(opened_message.decode("utf-8"))


def decrypt_content(
    form_private_key: str, encrypted_content: str
) -> Union[bytes, None]:
    try:
        [submission_public_key, nonce_encrypted] = encrypted_content.split(";")
        [nonce, encrypted] = list(
            map(lambda x: base64.b64decode(x), nonce_encrypted.split(":"))
        )
        private_key = PrivateKey(base64.b64decode(form_private_key))
        public_key = PublicKey(base64.b64decode(submission_public_key))
        box = Box(private_key, public_key)
        decrypted = box.decrypt(encrypted, nonce)
        return decrypted
    except CryptoError:
        logger.error(
            "Error decrypting, is your form_secret_key correct, or are you on the correct mode (staging/production)?"
        )
        return None


def retrieve_attachment_filenames(decrypted_content: Mapping[str, Any]) -> bool:
    return (
        decrypted_content["field_type"] == "attachment" and decrypted_content["answer"]
    )


def are_attachment_field_ids_valid(
    field_ids: List[str], filenames: Mapping[str, str]
) -> bool:
    return all(map(lambda field_id: field_id in filenames, field_ids))


def convert_encrypted_attachment_to_file_content(
    encrypted_attachment,
) -> EncryptedFileContent:
    logger.debug(
        "convert_encrypted_attachment_to_file_content.encrypted_attachment",
        encrypted_attachment,
    )
    return {
        "submission_public_key": encrypted_attachment["encryptedFile"][
            "submissionPublicKey"
        ],
        "nonce": encrypted_attachment["encryptedFile"]["nonce"],
        "binary": base64.b64decode(encrypted_attachment["encryptedFile"]["binary"]),
    }



import base64
import logging

from nacl.signing import VerifyKey

from formsg.schemas.verification import VerificationAuthenticateOptions
from formsg.util.parser import parse_verification_signature
from formsg.util.verification import format_to_base_string, is_signature_time_valid

logger = logging.getLogger(__name__)


class Verification(object):
    def __init__(
        self,
        verification_public_key: str,
        verification_secret_key: str,
        transaction_expiry: int,
    ):
        self.verification_public_key = verification_public_key
        self.verification_secret_key = verification_secret_key
        self.transaction_expiry = transaction_expiry

    """
      /**
   *  Verifies signature
   * @param {object} data
   * @param {string} data.signatureString
   * @param {number} data.submissionCreatedAt date in milliseconds
   * @param {string} data.fieldId
   * @param {string} data.answer
   * @param {string} data.publicKey
   */
    """

    def authenticate(
        self, verification_authenticate_options: VerificationAuthenticateOptions
    ):
        # FIXME: runtime checks at top level
        [signature_string, submission_created_at, field_id, answer] = [
            verification_authenticate_options["signatureString"],
            verification_authenticate_options["submissionCreatedAt"],
            verification_authenticate_options["fieldId"],
            verification_authenticate_options["answer"],
        ]
        try:
            verification_signature = parse_verification_signature(signature_string)
            [transaction_id, time_, form_id, signature] = [
                verification_signature["v"],
                verification_signature["t"],
                verification_signature["f"],
                verification_signature["s"],
            ]
            if not time_:
                raise Exception("Malformed signature string was passed into function")
            if is_signature_time_valid(
                time_, submission_created_at, self.transaction_expiry
            ):
                format_to_base_string_params = {
                    "transactionId": transaction_id,
                    "formId": form_id,
                    "fieldId": field_id,
                    "answer": answer,
                    "time": time_,
                }
                data = format_to_base_string(format_to_base_string_params)
                """
                return nacl.sign.detached.verify(
                    decodeUTF8(data),
                    decodeBase64(signature),
                    decodeBase64(this.verificationPublicKey)
                )
                """
                verify_key = VerifyKey(base64.b64decode(self.verification_public_key))
                return verify_key.verify(
                    smessage=data.encode("utf-8"),
                    signature=base64.b64decode(signature),
                )

            else:
                logger.info(
                    f'Signature was expired for signatureString="{signature_string}" signatureDate="{time_}" submissionCreatedAt="{submission_created_at}'
                )
                return False
        except Exception as e:
            print(e)
            return False
